@@ -1,41 +1,18 @@
-# import os.path, asyncio, json, hashlib, sys
-# import http.server, urllib.parse, webbrowser, secrets
 import weakref
 from enum import Enum
-# from json.decoder import JSONDecodeError
-# import datetime as dt
-# from dataclasses import dataclass
-from typing import Any, AsyncGenerator, Callable, Generic, TypeVar, cast
-
-# import base64
-# from hashlib import sha1
-# import hmac
-# import secrets
-
-# from urllib.parse import urlencode
-
-# from copy import deepcopy, copy
+from typing import Any, AsyncGenerator, Callable, Generator, Generic, TypeVar, cast, Awaitable
 
 from collections.abc import Coroutine
 
-import hashlib
-
 from aiohttp import ClientSession, ClientResponse
 from aiohttp.client_exceptions import ContentTypeError
 
-def nc_hash(x: str) -> str:
-    h = hashlib.new('ripemd160')
-    h.update(x.encode('utf8'))
-    return h.hexdigest()
+from copy import deepcopy
 
 T = TypeVar('T')
 U = TypeVar('U')
 
-Json = dict[str, 'Json'] | list['Json'] | str | int | float | None
-JsonMap = dict[str, Json]
-JsonScalar = str | int | float | None
 Coro = Coroutine[Any, Any, T]
-GenOrAsync = AsyncGenerator[T, list[T]]
 
 class AsyncLazy(Generic[T]):
     '''Does not accumulate any results unless awaited.'''
@@ -50,13 +27,13 @@ class AsyncLazy(Generic[T]):
     async def _items(self) -> list[T]:
         return [t async for t in self.gen]
 
-    def __await__(self):
+    def __await__(self) -> Generator[Any, None, list[T]]:
         return self._items().__await__()
 
     def map(self, f: Callable[[T], U]) -> 'AsyncTrans[U]':
         return AsyncTrans(self.gen, f)
 
-class AsyncTrans(AsyncLazy[Any], Generic[U]):
+class AsyncTrans(Generic[U], AsyncLazy[Any]):
     '''
     Does not accumulate any results unless awaited.
     Transforms the results of the generator using the mapping function.
@@ -71,6 +48,9 @@ class AsyncTrans(AsyncLazy[Any], Generic[U]):
     def __aiter__(self):
         return (self.mapping(t) async for t in self.gen)
 
+    def __await__(self) -> Generator[Any, None, list[U]]:
+        return self._items().__await__()
+
     async def _items(self) -> list[U]:
         return [u async for u in self]
 
@@ -86,48 +66,72 @@ class APIError(Exception):
     def __str__(self) -> str:
         return super().__str__() + F"\nStatus: {self.status}\nReason: {self.reason}"
 
-class EnumParam(Enum):
-    _params: list['EnumParam']
-    
-    def __str__(self) -> str:
-        return self.value
+class _EnumParams:
+    '''
+        Emulate an EnumParam for serialization into URL params.
+        Separate class and hidden since Enum's have special behavior.
+    '''
+    params: dict[str, set[str]]
 
-    def _get_title(self) -> str:
-        return self.__class__.__name__.lower()
+    def __init__(self):
+        self.params = {}
 
-    def __add__(self, other: 'EnumParam'):
-        if not hasattr(self, '_params'):
-            setattr(self, '_params', [self])
-        self._params.append(other)
-        return self
+    def __add__(self, other: 'EnumParam|_EnumParams') -> '_EnumParams':
+        new_instance = deepcopy(self)
+        match other:
+            case EnumParam():
+                other_items = [(other.get_title(), set([other.value]))]
+            case _EnumParams():
+                other_items = other.params.items()
+        for k, v in other_items:
+            if k not in new_instance.params:
+                new_instance.params[k] = set()
+            new_instance.params[k] |= v
+        return new_instance
 
     def to_dict(self, delimiter: str=',') -> dict[str, str]:
-        # TODO: If the value of a param is None, it should be omitted.
-        if not hasattr(self, '_params'):
-            setattr(self, '_params', [self])
-        params: dict[str, str] = {}
-        for param in self._params:
-            title = param._get_title()
-            if not title in params:
-                params[title] = param.value
-            else:
-                params[title] += delimiter+param.value
-        return params
+        '''
+            Convert packed parameters to a dictionary for use in a URL.
+        '''
+        return {
+            title: delimiter.join(values)
+            for title, values in self.params.items()
+        }
 
-async def api_err(response: ClientResponse, result: Json = None) -> APIError:
+class EnumParam(Enum):
+    '''
+        Collection of API url parameters which have only specific values.
+        Serializes to a dictionary for use in a URL.
+    '''
+    
+    def get_title(self) -> str:
+        return self.__class__.__name__[0].lower()+self.__class__.__name__[1:]
+
+    def __add__(self: T, other: 'Self|_EnumParams') -> T:
+        '''Collect with another parameter or set of parameters.'''
+        # return type is compatible with EnumParam for + and to_dict
+        return cast(EnumParam, _EnumParams() + self + other)
+    
+    def to_dict(self, delimiter: str=',') -> dict[str, str]:
+        '''
+            Convert packed parameters to a dictionary for use in a URL.
+        '''
+        return {
+            self.get_title(): self.value
+        }
+
+async def api_err(response: ClientResponse, result: Any = None) -> APIError:
     match result:
         case {'message': msg}:
             return APIError(response.status, msg)
         case _:
             return APIError(response.status, await response.text())
 
-class Paginator:
-    pass
-
 from .oauth2 import OAuth2User
 from .oauth1 import OAuth1User
 
 def convert_url_params(p: dict[str, Any]|None) -> dict[str, str]:
+    '''Excludes empty-valued parameters'''
     if p is None: return {}
     return {k: str(v) for k, v in p.items() if v is not None and v != ''}
 
@@ -173,7 +177,29 @@ class WebAPI:
         # is garbage collected.
         self._finalize = weakref.finalize(self, self.session._connector._close) # type: ignore ## reportPrivateUsage
 
+        # workaround for:
+        # https://github.com/aio-libs/aiohttp/issues/4324#issuecomment-733884349
+
+        # Replace the event loop destructor thing with one a wrapper which ignores
+        # this specific exception on windows.
+        import sys
+        if sys.platform.startswith("win"):
+            from asyncio.proactor_events import _ProactorBasePipeTransport # type: ignore
+
+            base_del = _ProactorBasePipeTransport.__del__
+            if not hasattr(base_del, '_once'):
+                def quiet_delete(*args, **kwargs): # type: ignore
+                    try:
+                        return base_del(*args, **kwargs) # type: ignore
+                    except RuntimeError as e:
+                        if str(e) != 'Event loop is closed':
+                            raise
+                quiet_delete._once = True # type: ignore
+
+                _ProactorBasePipeTransport.__del__ = quiet_delete
+
     def close(self):
+        '''Closes the http session with the API server. Should be automatic.'''
         self._finalize() 
 
     def _req(self, method: str, path: str, params: dict[str, Any]|None=None, json: Any=None, data: Any=None):
@@ -235,6 +261,7 @@ class WebAPI:
         path: str,
         params: dict[str, Any], # non-const
         limit: int|None) -> AsyncLazy[Any]:
+        '''Return an awaitable and async iterable over google-style paginated items'''
         return AsyncLazy(self._paginated(method, path, params, limit))
 
 # Endpoint = Callable[[WebAPI, T], Coro[U]]
